#ifndef Agent_h
#define Agent_h

#include <iostream>
#include <set>

// *** NOTE: "Cell" and "sort_danger" struct should not be given to students as a template ***
// Had no choice but to include them in Agent.hpp to resolve compilation issues 
struct Cell{
	int x;
	int y;
	int pit;
	int wumpus;
	int stench;
	int breeze;
	int visited;
};

// Sort by the danger likelihood
struct sort_danger {
	bool operator()(const Cell &c1, const Cell &c2) {
		return (c1.wumpus + 2 * c1.pit) < (c2.wumpus + 2 * c2.pit);
	}
};

class Agent{
    // An abstract class of agent
    // Need to overwrite two functions, getAction() and perceiveResult()
public:
    bool notShot;
	bool wumpusKilled;
	bool gotGold;
    bool gold;
    
    enum Action{TURN_LEFT, TURN_RIGHT, FORWARD, SHOOT, PICK_UP, CLIMB, NIL};
    enum Percept{STENCH, BREEZE, BUMP, GLITTER, SCREAM};
    
    Action lastAction;
    std::pair<int, int> facing;
    std::pair<int, int> position;
    
    Agent() {};
    
    virtual Action getAction() = 0;
    virtual void perceiveResult(std::set<Percept> percepts) = 0;
    
};


#endif /* Agent_h */
