/* Author: Minjae Wee	mwee@uci.edu */
// Wumpus World - Smart AI.hpp

#ifndef SmartAgent_hpp
#define SmartAgent_hpp

#include "Agent.hpp"
#include <algorithm>
#include <vector>
#include <string>
#include <map>

class SmartAgent : public Agent{
public:
    SmartAgent();
	bool checkCoordIn(std::pair<int, int> coord, std::vector<std::pair<int, int>> v);
	bool checkPerceptIn(Percept p);
	void visitCell();
	void updateWumpusPitProb(int r, int c, char wum_or_pit, int prob, bool setvalue);
	void placeExaminePercepts();
	void updateBoardSize();
	void planMove(std::pair<std::string, Cell> move);
	void dft(std::vector<std::pair<int, int>> safe_cells, std::pair<int, int> start);
	void DFTr(std::vector<std::pair<int, int>> G, std::pair<int, int> cell);
	std::pair<std::string, Cell> decideMove();
	std::vector<std::pair<int, int>> optimizeShortcut();
	std::vector<Cell> getPossibleMoves();
	std::vector<std::pair<int, int>> getAdjCells(std::pair<int, int> pos, std::vector<std::pair<int, int>> safe);
	std::vector<Action> actionsToTake(std::pair<int, int> start, std::pair<int, int> dest, std::pair<int, int> dir);
	std::string facingToString(std::pair<int,int> facing);
	std::string turnRight(std::string facing_temp);
	std::string turnLeft(std::string facing_temp);
	virtual Action getAction();
	virtual void perceiveResult(std::set<Percept> percepts);

private:
	int row;
	int col;
	int move_num;
	int saved_num;

	bool certain_wumpus;
	std::pair<int, int> saved_coord;
	std::pair<int, int> wumpus_pos;
	std::set<Agent::Percept> ps;
	std::vector<Action> move_proc;
	std::vector<std::vector<Cell>> board;
	std::vector<std::pair<int, int>> shortcut;
	std::map<std::pair<int, int>, bool> visited;
};

#endif /* SmartAgent_hpp */
