﻿#include "World.hpp"

// Constructor (Build the board as an array)
World::World(int r, int c)
    : row(r), column(c), running(false), gameOver(false)
{
	points = 0;
    board = new Tile*[r];
    for (int i=0; i<c; i++){
        board[i] = new Tile[c];
    }
}

// Deconstructor
World::~World(){
    for (int i=0; i<column; i++){
        delete[] board[i];
    }
    delete[] board;
}



// Genarate a legal world
void World::genWorld(){
    float pitRate = 0.2;
    
    // Generate pits
    for (int r=0; r<row; r++)
        for (int c=0; c<column; c++)
            board[r][c] = Tile();
    
    for (int r=0; r<row; r++)
        for (int c=0; c<column; c++)
            if ((r!=0)&&(c!=0))
                if (randomFloat() < pitRate)
                    board[r][c].pit = true;

    // Generate wumpus and gold
    int wr = randomInt(row); int wc = randomInt(column);
    int gr = randomInt(row); int gc = randomInt(column);
    
    // Make sure wumpus and gold are not in (0,0)
    while ((wr==0)&&(wc==0)){
        wr = randomInt(row);
        wc = randomInt(column);
    }
    
    while ((gr==0)&&(gc==0)){
        gr = randomInt(row);
        gc = randomInt(column);
    }
    
    board[wr][wc].wumpus = true;
    board[gr][gc].gold = true;
    
    // Update the board with stench and breeze
    updateBoard();
    
    // Initialize the agent
    board[0][0].agent = true;
}


// Update the board with stench and breeze
void World::updateBoard(){
    for (int r=0; r<row; r++){
        for (int c=0; c<column; c++){
            if (board[r][c].pit == true)
                updateBreeze(r,c);
            if (board[r][c].wumpus == true)
                updateStench(r,c);
        }
    }
}


// Check whether the input is in board
bool World::notOutOfBound(int r, int c){
    return (r>=0)&&(r<row)&&(c>=0)&&(c<column);
}

// Update breeze
void World::updateBreeze(int r, int c){
	if (notOutOfBound(r - 1, c))
		board[r - 1][c].breeze = true;
	if (notOutOfBound(r + 1, c))
		board[r + 1][c].breeze = true;
	if (notOutOfBound(r, c - 1))
		board[r][c - 1].breeze = true;
	if (notOutOfBound(r, c + 1))
		board[r][c + 1].breeze = true;
}

// Update stench
void World::updateStench(int r, int c){
	if (notOutOfBound(r, c))
		board[r][c].stench = true;
	if (notOutOfBound(r - 1, c))
		board[r - 1][c].stench = true;
	if (notOutOfBound(r + 1, c))
		board[r + 1][c].stench = true;
	if (notOutOfBound(r, c - 1))
		board[r][c - 1].stench = true;
	if (notOutOfBound(r, c + 1))
		board[r][c + 1].stench = true;
}




// Start game
void World::run(){
    running = true;
    
    // Initialize the agent
	agent = SmartAgent();
	//agent = AverageAgent();
	//agent = PoorAgent();
	//agent = DummyAgent();

	while (true){
		if (!turbo){
			char s;
			std::cout << "Continue? (if yes, type 'c' and press Enter): ";
			std::cin >> s;
			std::cout << "\n\n\n" << std::endl;

			if (s != 'c')
				continue;
		}
        // Get the action
		detectPercepts();
		Agent::Action action = agent.getAction();
        
        // Make sure ps is empty
        ps.clear();

		agent.lastAction = action;

        if (action==Agent::FORWARD){
            board[agent.position.first][agent.position.second].agent=false;
            agent.position = move(agent.position, agent.facing);
            board[agent.position.first][agent.position.second].agent=true;
			points--;
        }
        else if (action==Agent::TURN_RIGHT){
            if (agent.facing.first == 0){
                agent.facing = std::pair<int,int>(agent.facing.second, agent.facing.first);
            }
            else{
                agent.facing = std::pair<int,int>(agent.facing.second, -agent.facing.first);
            }
			points--;
        }
        else if (action==Agent::TURN_LEFT){
            if (agent.facing.second == 0){
                agent.facing = std::pair<int,int>(agent.facing.second, agent.facing.first);
            }
            else{
                agent.facing = std::pair<int,int>(-agent.facing.second, agent.facing.first);
            }
			points--;
        }
        else if (action==Agent::SHOOT){
            if (agent.notShot){
                std::pair<int,int> tile = move(agent.position, agent.facing);
                if (board[tile.first][tile.second].wumpus){
                    board[tile.first][tile.second].wumpus = false;
                    board[tile.first][tile.second].deadWumpus = true;
					ps.insert(Agent::SCREAM);
				}
			}
			agent.notShot = false;
			points -= 10;
        }
        else if (action==Agent::PICK_UP){
			points--;
            if (board[agent.position.first][agent.position.second].gold){
                board[agent.position.first][agent.position.second].gold = false;
                agent.gold = true;
            }
        }
        else if (action==Agent::CLIMB){
			points--;
            if ((agent.gold)&&(agent.position.first==0)&&(agent.position.second==0)){
                gameOver = true;
                running = false;
				if (agent.gotGold){
					points += 1000;
				}
				printBoard();
				printGameInfo();
				if (agent.gotGold){
					std::cout << "you win!" << std::endl;
				}
				else{
					std::cout << "Climbed out without the gold!" << std::endl;
				}
                break;
            }
        }

        // Game over if agent in pit or wumpus
        if ((board[agent.position.first][agent.position.second].pit == true) || (board[agent.position.first][agent.position.second].wumpus == true)){
            gameOver = true;
            running = false;
			points -= 1000;
			printBoard();
			printGameInfo();
            std::cout << "game over" << std::endl;
            break;
        }

		// Build the perception set and pass it to the agent
		detectPercepts();

		// Print out the board and details for debugging purpose
		printBoard();
		printGameInfo();
	}
}

void World::detectPercepts(){
	allPercepts(ps, agent.position, agent.facing);
	agent.perceiveResult(ps);
}

void World::printPerceptions(){
	std::cout << "Perceptions: ";
	printSet(ps);
}

void World::printScore(){
	std::cout << "Score: " << points << std::endl;
}

// Print out which direction the agent is currently facing
void World::printAgentDirection(){
	std::string direction;
	if (agent.facing.first == 1 && agent.facing.second == 0)
		direction = "DOWN";
	else if (agent.facing.first == -1 && agent.facing.second == 0)
		direction = "UP";
	else if (agent.facing.first == 0 && agent.facing.second == 1)
		direction = "RIGHT";
	else if (agent.facing.first == 0 && agent.facing.second == -1)
		direction = "LEFT";

	std::cout << "Agent Facing: " << direction << std::endl;
}

std::string World::getLastAction(Agent::Action a){
	if (a == Agent::TURN_LEFT){
		return "TURNED LEFT";
	}
	else if (a == Agent::TURN_RIGHT){
		return "TURNED RIGHT";
	}
	else if (a == Agent::FORWARD){
		return "MOVED FORWARD";
	}
	else if (a == Agent::SHOOT){
		return "SHOT THE ARROW";
	}
	else if (a == Agent::PICK_UP){
		return "PICK UP";
	}
	else if (a == Agent::CLIMB){
		return "CLIMB";
	}
	else if (a == Agent::NIL){
		return "NONE";
	}
}

void World::printLastAction(){
	std::cout << "Last Action: " << getLastAction(agent.lastAction) << std::endl;
}

void World::printGameInfo(){
	printScore();
	printPerceptions();
	printAgentDirection();
	printLastAction();
}

// Return the new position based on current position and facing (will not go out of bound)
std::pair<int,int> World::move(std::pair<int,int> p, std::pair<int,int> f){
    int newX;
    int newY;
    
    if (p.first+f.first > row-1){
        newX = row-1;
    }
    else if (p.first+f.first < 0){
        newX = 0;
    }
    else {
        newX = p.first+f.first;
    }
    
    if (p.second+f.second > column-1){
        newY = column-1;
    }
    else if (p.second+f.second < 0){
        newY = 0;
    }
    else {
        newY = p.second+f.second;
    }
    
    return std::pair<int,int>(newX, newY);
}

// Build the perception set
void World::allPercepts(std::set<Agent::Percept>& s, std::pair<int,int> p, std::pair<int,int> f){
    if (board[p.first][p.second].stench){
        s.insert(Agent::STENCH);
    }
    if (board[p.first][p.second].breeze){
        s.insert(Agent::BREEZE);
    }
    if (board[p.first][p.second].gold){
        s.insert(Agent::GLITTER);
    }
    if (p==move(p,f)){
        s.insert(Agent::BUMP);
    }
}

// Print for debug
void World::printBoard(){
	std::cout << std::endl;
    for (int r=0; r<row; r++){
        for (int c=0; c<column; c++){
            std::string result = board[r][c].stringTile();
            std::cout << result;
            for (int i=0; i<30-result.length(); i++){
                std::cout << " ";
            }
        }
		std::cout << '\n' << std::endl;
    }
}


std::string World::getPercept(Agent::Percept p){
    if (p == Agent::STENCH){
        return "stench";
    }
    else if (p == Agent::BREEZE){
        return "breeze";
    }
    else if (p == Agent::BUMP){
        return "bump";
    }
    else if (p == Agent::GLITTER){
        return "glitter";
    }
    else if (p == Agent::SCREAM){
        return "scream";
    }
    else {
        return "I have no idea what the heck is going on.";
    }
}

void World::printSet(std::set<Agent::Percept>& s){
    for (auto i=s.begin(); i!=s.end(); i++){
        std::cout << getPercept(*i) << " ";
    }
    std::cout << std::endl;
}